@@ -275,6 +275,73 @@ static void cc_miller(element_t res, mpz_t q, point_t P,
     element_clear(e1);
 }
 
+static void point_proj_double(point_t R, element_t Rz, point_t P, element_t z)
+{
+    common_curve_ptr cc = P->curve->data;
+    element_t e0, e1, e2, e3;
+    element_init(e0, z->field);
+    element_init(e1, z->field);
+    element_init(e2, z->field);
+    element_init(e3, z->field);
+    element_ptr x = P->x;
+    element_ptr y = P->y;
+    //e0 = 3x^2 + (cc->a) z^4
+    element_square(e0, x);
+    element_mul_si(e0, e0, 3);
+    element_square(e1, z);
+    element_square(e1, e1);
+    element_mul(e1, e1, cc->a);
+    element_add(e0, e0, e1);
+
+    //z_out = 2 y z
+    element_mul(Rz, y, z);
+    element_mul_si(Rz, Rz, 2);
+
+    //e1 = 4 x y^2
+    element_square(e2, y);
+    element_mul(e1, x, e2);
+    element_mul_si(e1, e1, 4);
+
+    //x_out = e0^2 - 2 e1
+    element_mul_si(e3, e1, 2);
+    element_square(R->x, e0);
+    element_sub(R->x, R->x, e3);
+
+    //e2 = 8y^4
+    element_square(e2, e2);
+    element_mul_si(e2, e2, 8);
+
+    //y_out = e0(e1 - x_out) - e2
+    element_sub(e1, e1, R->x);
+    element_mul(e0, e0, e1);
+    element_sub(R->y, e0, e2);
+
+    R->inf_flag = 0;
+
+    element_clear(e0);
+    element_clear(e1);
+    element_clear(e2);
+    element_clear(e3);
+}
+
+//convert P from weighted projective (Jacobian) to affine
+//i.e. (X, Y, Z) --> (X/Z^2, Y/Z^3)
+//also sets z to 1
+static void point_to_affine(point_ptr P, element_t z)
+{
+    element_t e0;
+
+    element_init(e0, z->field);
+    element_invert(z, z);
+    element_square(e0, z);
+    element_mul(P->x, P->x, e0);
+    element_mul(e0, e0, z);
+    element_mul(P->y, P->y, e0);
+    element_set1(z);
+
+    element_clear(e0);
+}
+
 static void cc_tatepower(element_ptr out, element_ptr in, pairing_t pairing)
 {
     mnt_pairing_data_ptr p = pairing->data;
@@ -409,6 +476,7 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     common_curve_ptr cc = P->curve->data;
     element_t t0;
     element_t e0, e1;
+    element_t z;
 
     void do_vertical(void)
     {
@@ -518,6 +586,8 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_init(t0, P->curve->field);
     element_init(e0, res->field);
     element_init(e1, res->field);
+    element_init(z, a->field);
+    element_set1(z);
 
     element_init(v, res->field);
     point_init(Z, P->curve);
@@ -548,6 +618,7 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_clear(t0);
     element_clear(e0);
     element_clear(e1);
+    element_clear(z);
 }
 
 static void cc_tatepower_even_k(element_ptr out, element_ptr in, pairing_t pairing)
