@@ -42,27 +42,32 @@ void e_param_gen(e_param_t p, int rbits, int qbits)
     element_t j;
     int found = 0;
 
-    p->exp2 = rbits;
     mpz_init(n);
 
     do {
 	int i;
 	mpz_set_ui(r, 0);
+
+	if (rand() % 2) {
+	    p->exp2 = rbits - 1;
+	    p->sign1 = 1;
+	} else {
+	    p->exp2 = rbits;
+	    p->sign1 = -1;
+	}
 	mpz_setbit(r, p->exp2);
 
+	p->exp1 = (rand() % (p->exp2 - 1)) + 1;
 	//use q as a temp variable
 	mpz_set_ui(q, 0);
-	do {
-	    p->exp1 = rand() % rbits;
-	} while (!p->exp1);
 	mpz_setbit(q, p->exp1);
-	if (rand() % 2) {
-	    p->sign1 = 1;
+
+	if (p->sign1 > 0) {
 	    mpz_add(r, r, q);
 	} else {
-	    p->sign1 = -1;
 	    mpz_sub(r, r, q);
 	}
+
 	if (rand() % 2) {
 	    p->sign0 = 1;
 	    mpz_add_ui(r, r, 1);
@@ -74,7 +79,7 @@ void e_param_gen(e_param_t p, int rbits, int qbits)
 	for (i=0; i<10; i++) {
 	    //use q as a temp variable
 	    mpz_set_ui(q, 0);
-	    mpz_setbit(q, hbits);
+	    mpz_setbit(q, hbits + 1);
 	    pbc_mpz_random(h, q);
 	    mpz_mul(h, h, h);
 	    mpz_mul_ui(h, h, 28);
@@ -162,7 +167,9 @@ void e_param_inp_str(e_param_ptr p, FILE *stream)
 }
 
 static void e_miller(element_t res, point_t P,
-	element_ptr Qx, element_ptr Qy, e_pairing_data_ptr p)
+	element_ptr numx, element_ptr numy,
+	element_ptr denomx, element_ptr denomy,
+	e_pairing_data_ptr p)
 {
     //collate divisions
     int n;
@@ -175,16 +182,19 @@ static void e_miller(element_t res, point_t P,
     element_t e0, e1;
     int i;
 
-    void do_vertical(element_t e, point_ptr A)
+    void do_vertical(element_t e, element_t edenom, point_ptr A)
     {
 	if (point_is_inf(A)) {
 	    return;
 	}
-	element_sub(e0, Qx, A->x);
+	element_sub(e0, numx, A->x);
 	element_mul(e, e, e0);
+
+	element_sub(e0, denomx, A->x);
+	element_mul(edenom, edenom, e0);
     }
 
-    void do_tangent(element_t e, point_ptr A)
+    void do_tangent(element_t e, element_t edenom, point_ptr A)
     {
 	//a = -slope_tangent(A.x, A.y);
 	//b = 1;
@@ -205,7 +215,7 @@ static void e_miller(element_t res, point_t P,
 
 	if (element_is0(Ay)) {
 	    //order 2
-	    do_vertical(e, A);
+	    do_vertical(e, edenom, A);
 	    return;
 	}
 
@@ -221,14 +231,20 @@ static void e_miller(element_t res, point_t P,
 	element_add(c, c, t0);
 	element_neg(c, c);
 
-	element_mul(e0, a, Qx);
-	element_mul(e1, b, Qy);
+	element_mul(e0, a, numx);
+	element_mul(e1, b, numy);
 	element_add(e0, e0, e1);
 	element_add(e0, e0, c);
 	element_mul(e, e, e0);
+
+	element_mul(e0, a, denomx);
+	element_mul(e1, b, denomy);
+	element_add(e0, e0, e1);
+	element_add(e0, e0, c);
+	element_mul(edenom, edenom, e0);
     }
 
-    void do_line(element_ptr e, point_ptr A, point_ptr B)
+    void do_line(element_ptr e, element_ptr edenom, point_ptr A, point_ptr B)
     {
 	//a = -(B.y - A.y) / (B.x - A.x);
 	//b = 1;
@@ -242,16 +258,16 @@ static void e_miller(element_t res, point_t P,
 
 	//we assume B is never O
 	if (point_is_inf(A)) {
-	    do_vertical(e, A);
+	    do_vertical(e, edenom, A);
 	    return;
 	}
 
 	if (!element_cmp(Ax, Bx)) {
 	    if (!element_cmp(Ay, By)) {
-		do_tangent(e, A);
+		do_tangent(e, edenom, A);
 		return;
 	    }
-	    do_vertical(e, A);
+	    do_vertical(e, edenom, A);
 	    return;
 	}
 
@@ -262,11 +278,17 @@ static void e_miller(element_t res, point_t P,
 	element_add(c, c, t0);
 	element_neg(c, c);
 
-	element_mul(e0, a, Qx);
-	element_mul(e1, b, Qy);
+	element_mul(e0, a, numx);
+	element_mul(e1, b, numy);
 	element_add(e0, e0, e1);
 	element_add(e0, e0, c);
 	element_mul(e, e, e0);
+
+	element_mul(e0, a, denomx);
+	element_mul(e1, b, denomy);
+	element_add(e0, e0, e1);
+	element_add(e0, e0, c);
+	element_mul(edenom, edenom, e0);
     }
 
     element_init(a, P->curve->field);
@@ -294,14 +316,14 @@ static void e_miller(element_t res, point_t P,
     for (i=0; i<n; i++) {
 	element_square(v, v);
 	element_square(vd, vd);
-	do_tangent(v, Z);
+	do_tangent(v, vd, Z);
 	point_double(Z, Z);
-	do_vertical(vd, Z);
+	do_vertical(vd, v, Z);
     }
     if (p->sign1 < 0) {
 	element_set(v1, vd);
 	element_set(vd1, v);
-	do_vertical(vd1, Z);
+	do_vertical(vd1, v1, Z);
 	point_neg(Z1, Z);
     } else {
 	element_set(v1, v);
@@ -312,18 +334,18 @@ static void e_miller(element_t res, point_t P,
     for (; i<n; i++) {
 	element_square(v, v);
 	element_square(vd, vd);
-	do_tangent(v, Z);
+	do_tangent(v, vd, Z);
 	point_double(Z, Z);
-	do_vertical(vd, Z);
+	do_vertical(vd, v, Z);
     }
     element_mul(v, v, v1);
     element_mul(vd, vd, vd1);
-    do_line(v, Z, Z1);
+    do_line(v, vd, Z, Z1);
     point_add(Z, Z, Z1);
-    do_vertical(vd, Z);
+    do_vertical(vd, v, Z);
 
     if (p->sign0 > 0) {
-	do_vertical(v, P);
+	do_vertical(v, vd, P);
     }
 
     element_invert(vd, vd);
@@ -510,12 +532,11 @@ static void e_pairing(element_ptr out, element_ptr in1, element_ptr in2,
     if (0) {
 	cc_miller(out, pairing->r, in1->data, QR->x, QR->y);
 	cc_miller(e0, pairing->r, in1->data, R->x, R->y);
+	element_invert(e0, e0);
+	element_mul(out, out, e0);
     } else {
-	e_miller(out, in1->data, QR->x, QR->y, p);
-	e_miller(e0, in1->data, R->x, R->y, p);
+	e_miller(out, in1->data, QR->x, QR->y, R->x, R->y, p);
     }
-    element_invert(e0, e0);
-    element_mul(out, out, e0);
     element_pow(out, out, p->tateexp);
     point_clear(R);
     point_clear(QR);
