@@ -14,8 +14,8 @@
 struct field_s;
 
 struct element_s {
-    void *data;
     struct field_s *field;
+    void *data;
 };
 typedef struct element_s *element_ptr;
 typedef struct element_s element_t[1];
@@ -52,8 +52,8 @@ struct field_s {
     int fixed_length_in_bytes; //length of an element in bytes; -1 for variable
     mpz_t order; //-1 for infinite order
     element_ptr nqr; //nonquadratic residue
-    void *data;
     void (*to_mpz)(mpz_ptr, element_ptr); //does this make sense for all fields?
+    void *data;
 };
 typedef struct field_s *field_ptr;
 typedef struct field_s field_t[1];
@@ -206,8 +206,6 @@ static inline int element_length_in_bytes(element_ptr a)
     }
 }
 
-void field_init_fp(field_ptr f, mpz_t prime);
-
 void element_pow2(element_ptr x, element_ptr a1, mpz_ptr n1,
                                  element_ptr a2, mpz_ptr n2);
 void element_pow3(element_ptr x, element_ptr a1, mpz_ptr n1,
@@ -233,4 +231,12 @@ static inline int mpz_is0(mpz_ptr z)
     return !mpz_sgn(z);
     //return !mpz_cmp_ui(z, 0);
 }
+
+void field_init_naive_fp(field_ptr f, mpz_t prime);
+void field_init_slow_fp(field_ptr f, mpz_t prime);
+void field_init_fast_fp(field_ptr f, mpz_t prime);
+
+static inline void field_init_fp(field_ptr f, mpz_t prime) {
+    field_init_naive_fp(f, prime);
+}
 #endif //FIELD_H
