@@ -175,6 +175,32 @@ static void poly_neg(element_ptr f, element_ptr g)
     }
 }
 
+static void poly_mul_mpz(element_ptr f, element_ptr g, mpz_ptr z)
+{
+    poly_element_ptr pf = f->data;
+    poly_element_ptr pg = g->data;
+    int i, n;
+
+    n = pg->coeff->count;
+    poly_alloc(f, n);
+    for (i=0; i<n; i++) {
+	element_mul_mpz(pf->coeff->item[i], pg->coeff->item[i], z);
+    }
+}
+
+static void poly_mul_si(element_ptr f, element_ptr g, signed long int z)
+{
+    poly_element_ptr pf = f->data;
+    poly_element_ptr pg = g->data;
+    int i, n;
+
+    n = pg->coeff->count;
+    poly_alloc(f, n);
+    for (i=0; i<n; i++) {
+	element_mul_si(pf->coeff->item[i], pg->coeff->item[i], z);
+    }
+}
+
 static void poly_mul(element_ptr r, element_ptr f, element_ptr g)
 {
     poly_element_ptr pprod;
@@ -510,6 +536,8 @@ void field_init_poly(field_ptr f, field_ptr base_field)
     f->sub = poly_sub;
     f->neg = poly_neg;
     f->mul = poly_mul;
+    f->mul_mpz = poly_mul_mpz;
+    f->mul_si = poly_mul_si;
     f->cmp = poly_cmp;
 }
 
@@ -780,6 +808,26 @@ static void polymod_sub(element_t r, element_t e, element_t f)
     }
 }
 
+static void polymod_mul_mpz(element_t e, element_t f, mpz_ptr z)
+{
+    polymod_field_data_ptr p = e->field->data;
+    element_t *dst = e->data, *src = f->data;
+    int i, n = p->n;
+    for (i=0; i<n; i++) {
+	element_mul_mpz(dst[i], src[i], z);
+    }
+}
+
+static void polymod_mul_si(element_t e, element_t f, signed long int z)
+{
+    polymod_field_data_ptr p = e->field->data;
+    element_t *dst = e->data, *src = f->data;
+    int i, n = p->n;
+    for (i=0; i<n; i++) {
+	element_mul_si(dst[i], src[i], z);
+    }
+}
+
 static void polymod_mul(element_ptr res, element_ptr e, element_ptr f)
 {
     polymod_field_data_ptr p = res->field->data;
@@ -793,15 +841,56 @@ static void polymod_mul(element_ptr res, element_ptr e, element_ptr f)
     element_init(c0, p->field);
 
     for (i=0; i<n; i++) {
-	//TODO: more efficient logic
-	for (j=0; j<n; j++) {
+	int ni = n - i;
+	for (j=0; j<ni; j++) {
 	    element_mul(c0, s1[i], s2[j]);
-	    if (i + j < n) {
-		element_add(dst[i + j], dst[i + j], c0);
-	    } else {
-		polymod_const_mul(p0, c0, p->xpwr[i + j - n]);
-		element_add(prod, prod, p0);
-	    }
+	    element_add(dst[i + j], dst[i + j], c0);
+	}
+	for (;j<n; j++) {
+	    element_mul(c0, s1[i], s2[j]);
+	    polymod_const_mul(p0, c0, p->xpwr[j - ni]);
+	    element_add(prod, prod, p0);
+	}
+    }
+
+    element_set(res, prod);
+    element_clear(prod);
+    element_clear(p0);
+    element_clear(c0);
+}
+
+static void polymod_square(element_ptr res, element_ptr e)
+{
+    polymod_field_data_ptr p = res->field->data;
+    element_t *dst, *src = e->data;
+    element_t prod, p0, c0;
+    int i, j, n = p->n;
+
+    element_init(prod, res->field);
+    dst = prod->data;
+    element_init(p0, res->field);
+    element_init(c0, p->field);
+
+    for (i=0; i<n; i++) {
+	int twicei = 2 * i;
+	element_square(c0, src[i]);
+	if (twicei < n) {
+	    element_add(dst[twicei], dst[twicei], c0);
+	} else {
+	    polymod_const_mul(p0, c0, p->xpwr[twicei - n]);
+	    element_add(prod, prod, p0);
+	}
+
+	for (j=i+1; j<n-i; j++) {
+	    element_mul(c0, src[i], src[j]);
+	    element_add(c0, c0, c0);
+	    element_add(dst[i + j], dst[i + j], c0);
+	}
+	for (;j<n; j++) {
+	    element_mul(c0, src[i], src[j]);
+	    element_add(c0, c0, c0);
+	    polymod_const_mul(p0, c0, p->xpwr[i + j - n]);
+	    element_add(prod, prod, p0);
 	}
     }
 
@@ -939,6 +1028,9 @@ void field_init_polymod(field_ptr f, element_ptr poly)
     f->set1 = polymod_set1;
     f->cmp = polymod_cmp;
     f->mul = polymod_mul;
+    f->mul_mpz = polymod_mul_mpz;
+    f->mul_si = polymod_mul_si;
+    f->square = polymod_square;
     f->random = polymod_random;
     f->from_hash = polymod_from_hash;
     f->invert = polymod_invert;
